%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [alpha fk dual]=ncssvm_wb(x, y, hkernel, kernelparam, lambda, gamma, alpha)
% function [alpha fk dual DD]=ncssvm_wb(x, y, hkernel, kernelparam, lambda, gamma, alpha)
%   x               N x p data
%   y               N x 1 label
%   hkernel         kernel function handle
%   kernelparam     kernel parameter
%   lambda          regularization parameter
%   gamma           M x 1 [0,1] cost asymmetries of interest
%   alpha           N x M initial alpha's at the cost asymmetry gamma
% 
%   alpha           N x M parameters
%   fk              decision function value ; > 0 positive ; < 0 negative
%   dual            objective function value
% 
%   nested cost sensitive support vector machine
%   find nested decision rules at multiple cost asymmetry at the same time
% 
% Gyemin Lee, University of Michigan
% 
tol1 = 1e-9;
tol2 = 1e-6;
tol = tol1;
EPS = 1e-10;

N = size(x,1);
M = length(gamma);
K = hkernel(x,x,kernelparam);
K = (K+K')/2;
Kt = diag(y)*K*diag(y);

loopnum = 0;
tolloopnum1 = 15000;
maxloopnum = 30000;
bStop = false;
prevD = Inf;
% DD = zeros(maxloopnum,1);
% pplst = zeros(maxloopnum,1);
% errlst = zeros(maxloopnum,1);

% initialization
if nargin < 7
    alpha = repmat(y==-1,1,M)+y*gamma';
end

prevAlpha = alpha;
preverr = Inf;
prevpp = 0;
bpp = 0;

while (~bStop)
    loopnum = loopnum + 1;
    
    % select a data point that violates the KKT condition most
    gim = Kt*alpha/lambda -1;
    [pp err] = selection(y, gamma, alpha, gim, bpp);
%     pplst(loopnum) = pp;
%     errlst(loopnum) = err;
    
    % solve the subproblem
    anew = alpha(pp,:)' - lambda*gim(pp,:)'/K(pp,pp);
    alpha(pp,:) = subproblem(y(pp), [N M], gamma, anew, alpha(pp,:)')';

    % dual value
    D = sum(diag(alpha'*Kt*alpha))/(2*lambda) - sum(alpha(:));
    if (D > prevD + tol)
        fprintf('WARNING: dual value increased\n');
        alpha = prevAlpha;
        bStop = true;
    else
        if (err < tol || loopnum == maxloopnum)
            bStop = true;
        end
        if (loopnum == tolloopnum1)
            tol = tol2;
        end

        if (abs(preverr-err)<EPS) && (max(abs(prevAlpha(:)-alpha(:)))<EPS)
            if (prevpp==pp)
                bpp = 1;
            elseif bpp
                fprintf('WARNING: ncssvm_wb.m : sturated\n');
                bpp = bpp+1;
                if bpp>=N
                    bStop = true;
                end
            end
        else
            bpp = 0;
        end
        prevD = D;
        prevAlpha = alpha;
        prevpp = pp;
        preverr = err;
%         DD(loopnum) = D;
    end
end

fprintf('ncs_svm: loop %d \t D: %e \t err: %e', loopnum, D, err);
if (err>tol), fprintf('<-----'); end
fprintf('\n');

fk = K*(alpha.*repmat(y,1,M));
dual = sum(diag(alpha'*Kt*alpha))/(2*lambda) - sum(alpha(:));
% DD = DD(1:loopnum);


function [sel_idx sel_err] = selection(y, gamma, alpha, gim, bpp)

tol = 1e-12;

[N M] = size(alpha);
Ip = y== 1;
Im = y==-1;

vim = zeros(N,M);
ferr = zeros(N,M);

% I+
mm = M;
idx1 = alpha(:,mm) < gamma(mm) - tol;
idx3 = alpha(:,mm-1) < alpha(:,mm) - tol;

ferr(Ip & ~idx1,mm) = gim(Ip & ~idx1,mm).*(gim(Ip & ~idx1,mm)>0);
vim( Ip &  idx1 & ~idx3,mm) =     gim(Ip &  idx1 & ~idx3,mm);
ferr(Ip &  idx1 &  idx3,mm) = abs(gim(Ip &  idx1 &  idx3,mm));

for mm=M-1:-1:1
    idx1 = alpha(:,mm) < gamma(mm) - tol;
    idx2 = alpha(:,mm) < alpha(:,mm+1) - tol;
    if (mm>1)
        idx3 = alpha(:,mm-1) < alpha(:,mm) - tol; 
    else
        idx3 = tol < alpha(:,mm);
    end
    if (gamma(mm)==0)
        break
    end
    
    ferr(Ip & ~idx1 &  idx2,mm) = gim(Ip & ~idx1 &  idx2,mm);
    ferr(Ip & ~idx1 & ~idx2,mm) = gim(Ip & ~idx1 & ~idx2,mm) ...
        + vim(Ip & ~idx1 & ~idx2,mm+1);
    ferr(Ip & ~idx1,mm) = ferr(Ip & ~idx1,mm).*(ferr(Ip & ~idx1,mm)>0);
    
    vim( Ip &  idx1 &  idx2 & ~idx3,mm) = gim(Ip &  idx1 &  idx2 & ~idx3,mm);
    vim( Ip &  idx1 & ~idx2 & ~idx3,mm) = gim(Ip &  idx1 & ~idx2 & ~idx3,mm) ...
        + vim(Ip &  idx1 & ~idx2 & ~idx3,mm+1);
    
    ferr(Ip &  idx1 &  idx2 &  idx3,mm) = abs(gim(Ip &  idx1 &  idx2 &  idx3,mm));
    ferr(Ip &  idx1 & ~idx2 &  idx3,mm) = abs(gim(Ip &  idx1 & ~idx2 &  idx3,mm) ...
        + vim(Ip &  idx1 & ~idx2 &  idx3,mm+1));
end

% I-
mm = 1;
idx1 = alpha(:,mm) < 1-gamma(mm) - tol;
idx3 = alpha(:,mm+1) < alpha(:,mm) - tol;

ferr(Im & ~idx1,mm) = gim(Im & ~idx1,mm).*(gim(Im & ~idx1,mm)>0);
vim( Im &  idx1 & ~idx3,mm) =     gim(Im &  idx1 & ~idx3,mm);
ferr(Im &  idx1 &  idx3,mm) = abs(gim(Im &  idx1 &  idx3,mm));

for mm=2:M
    idx1 = alpha(:,mm) < 1-gamma(mm) - tol;
    idx2 = alpha(:,mm) < alpha(:,mm-1) - tol;
    if (mm<M)
        idx3 = alpha(:,mm+1) < alpha(:,mm) - tol;
    else
        idx3 = tol < alpha(:,mm);
    end
    if (gamma(mm)==1)
        break
    end
    
    ferr(Im & ~idx1 &  idx2,mm) = gim(Im & ~idx1 &  idx2,mm);
    ferr(Im & ~idx1 & ~idx2,mm) = gim(Im & ~idx1 & ~idx2,mm) ...
        + vim(Im & ~idx1 & ~idx2,mm-1);
    ferr(Im & ~idx1,mm) = ferr(Im & ~idx1,mm).*(ferr(Im & ~idx1,mm)>0);
    
    vim( Im &  idx1 &  idx2 & ~idx3,mm) = gim(Im &  idx1 &  idx2 & ~idx3,mm);
    vim( Im &  idx1 & ~idx2 & ~idx3,mm) = gim(Im &  idx1 & ~idx2 & ~idx3,mm) ...
        + vim(Im &  idx1 & ~idx2 & ~idx3,mm-1);
    
    ferr(Im &  idx1 &  idx2 &  idx3,mm) = abs(gim(Im &  idx1 &  idx2 &  idx3,mm));
    ferr(Im &  idx1 & ~idx2 &  idx3,mm) = abs(gim(Im &  idx1 & ~idx2 &  idx3,mm) ...
        + vim(Im &  idx1 & ~idx2 &  idx3,mm-1));
end

verrsum = -sum(vim.*(vim<0),2);
ferrsum = sum(ferr,2);

[errlst, erridx] = sort(verrsum+ferrsum, 'descend');
sel_err = errlst(1);
sel_idx = erridx(1+bpp);


function alpha_new = subproblem(y, size_alpha, gamma, anew, aold)

N = size_alpha(1);
M = size_alpha(2);
aold = (y==-1)+y*gamma;

options = optimset('LargeScale', 'off', 'Display', 'off');
H = eye(M);
f = -anew;
A = -H + [zeros(1,M) ; H(1:M-1,:)];
if (y==-1), A = A.'; end
ub = (y==-1)+y*gamma;

[alpha_new,D,exitflag]=quadprog(H, f, A, zeros(M,1), [], [], [], ub, aold, options);
